#include <vector>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

//#define DEBUG

using namespace std;

int addrange(unsigned int add_start, unsigned int add_end);
int validate(unsigned int add_start, unsigned int add_end);

std::vector<unsigned int> a;
std::vector<unsigned int> b;
std::vector<unsigned int> linear_base_addresses;
std::vector<unsigned int> extended_segment_addresses;
std::vector<unsigned int> ip_register_values;
std::vector<unsigned int> cs_register_values;
std::vector<unsigned int> eip_values;

/*********************************************
Function Name:addrange(int add_start, int add_end)
Description: adds arrange to vectors for verification
Input: start address, end address
output: 1 if success, -1 if error
*********************************************/
int addrange(unsigned int add_start, unsigned int add_end)
{

	if(a.size() != b.size()){	// check if both vector size are same
		fprintf( stderr , "Mismathch in size of a and b\n" );
		return -1;
	}

	if(validate(add_start, add_end) == 1){
		a.push_back(add_start);
		b.push_back(add_end);
	}
	else{
		fprintf( stderr , "Invalid address range , start=%x, end=%x\n",add_start , add_end );
		return -1;
	}

	return 1;
}


/*********************************************
Function Name:validate(int add_start, int add_end)
Description: verifies if address range is valid
Input: start address, end address
output: 1 if success, -1, -2 if error
*********************************************/
int validate(unsigned int add_start, unsigned int add_end)
{
	unsigned int i=0;
	if(add_start >= add_end){
		fprintf( stderr , "End address should be greater then start address \n" );
		return -1;
	}

	for( i=0 ; i<a.size() ; ++i){
		if(add_start < a[i] && add_end < a[i])
			continue;
		else if (add_start > b[i] && add_end > b[i])
			continue;
		else{
			fprintf(stderr,"Overlapping address range\n");
			return -2;
		}
	}
	return 1;
}


/*******************************************************************************
// Function Name: read_char()
//Description: apart from the 1st colon and the End Of Line (EOL), no other characters
//than 0..9 and A..F are allowed in a record. Interpretation of a record
//should be case insensitive, so it does not matter if you use a..f or A..F.
//input: input_file and record_number
//output: 
// -1   --> Invalid Case
// -2   --> End of File
// ff   --> new line char
// 0-f  --> for 0-9, a-f or A-F
//
*******************************************************************************/
//int read_char( int input_file , unsigned int record_number ) {
int read_char( int input_file) {
	int byteread=0;
	char ch='\0';

	byteread = read( input_file , &ch , 1 );
	if(byteread == 0) {
		//printf("End of file\n");
		return -2;
	}
	if( (ch >= 'A') && (ch <= 'F') )
		ch = ch - 55;
	else if( (ch >= 'a') && (ch <= 'f') )
		ch = ch - 87;
	else if( (ch >= '0') && (ch <= '9') )
		ch = ch - 48;
	else if( ch == '\n' )
		return 0xff;
	else{
		//fprintf( stderr , "Invalid character in record number %u\n", record_number );
		return -1;
	}

	return (int)ch;
}
/*******************************************************************************
// Function Name:read_byte()
//Read two characters and return a byte
//Input: input_file , record_number
//Output:
// -1  --> for invalid character
// -2  --> for end of file
// xx  --> valid byte
****************************************************************************/
int read_byte( int input_file , unsigned int record_number ) {
	int byte=0;
	int temp=0;

	byte = read_char( input_file);
	if( ( byte == 0xff ) || ( byte == -1 ) ){
		fprintf( stderr , "Invalid character in record number %u\n" , record_number );
		return -1;
	}
	if ( byte == -2 )
		return -2;

	temp = read_char( input_file );
	if( ( byte == 0xff ) || ( byte == -1 ) ){
		fprintf( stderr , "Invalid character in record number %u\n" , record_number );
		return -1;
	}
	if ( byte == -2 )
		return -2;

	byte = ( byte<<4 ) | temp;

	return byte;
}
/*******************************************************************************
// Function Name:parse_newline_character( int input_file , unsigned int record_number )
Description: validate new line character
Input: input_file  record_number
output: error if new line is not \r\n or \n

********************************************************************************/

int parse_newline_character( int input_file , unsigned int record_number ) {

	int byteread;
	char ch;

	//Read carriage return
	byteread = read( input_file , &ch , 1 );
	if(byteread == 0) {
		printf("End of file\n");
 		return -2;
	}
	if( ch == '\r' ){
		byteread = read( input_file , &ch , 1 );
		if(byteread == 0) {
			//printf("End of file\n");
 			return -2;
		}
		if( ch == '\n' )
			return 1;
		else {
			fprintf( stderr , " Invalid character '%d' in record number %u\n", ch , record_number );
			return -1;
		}
	}
	else if( ch == '\n' )
		return 1;
	else {
		fprintf( stderr , " Invalid character in record number %u\n", record_number );
		return -1;
	}

}

/*******************************************************************************
// Function Name:parse_file(int input_file)
Description: parses the intel hex file
Input: input file handler
output: information about file
********************************************************************************/

unsigned int es_address=0 , lb_address = 0;
unsigned int cs_register , ip_register , eip=0;
char hex_type=0;
int record_number=1;

int parse_file(int input_file){
	int byteread=0;
	char ch='\0';
	char record_end_flag=0;
	int i=0;
	unsigned int old_address=0 , new_address=0;


	while(1){
		int byte_count=0, address=0, record_type=0, check_sum=0 , temp=0, temp1=0 , temp2=0;
		unsigned int absolute_address=0;


		while( 1 ) {
			//check in between new line characters and read valid identifier
			byteread = read( input_file , &ch , 1 );
			if(byteread == 0) {
				printf("End of file\n");
		 		return -2;
			}
			if( ch == '\r' ){
				byteread = read( input_file , &ch , 1 );
				if(byteread == 0) {
					printf("End of file\n");
		 			return -2;
				}
				if( ch == '\n' )
					continue;
				else
					break;
			}
			else if( ch == '\n' )
				continue;
			else
				break;
		}
		
		if( ch != ':' ){
			fprintf( stderr , "Identifier of record is not valid in line number %d\n", record_number );
			return -1;
		}
		#ifdef DEBUG
		printf( "Record Identifier = %c\n" , ch);
		#endif


		//Read two char - Byte count
		byte_count = read_byte( input_file , record_number );
		if( byte_count == -1)
			return -1;
		check_sum = byte_count;
		#ifdef DEBUG
		printf("Byte Count = %d\n" , byte_count );
		#endif


		//Read four char - Address
		address = read_byte( input_file , record_number );
		if( address == -1)
			return -1;
		check_sum = check_sum + address;
		temp = read_byte( input_file , record_number );
		if( temp == -1 )
			return -1;
		check_sum = check_sum + temp;
		address = ( address << 8 ) | temp;
		#ifdef DEBUG
		printf( "Address = %x\n" , address );
		#endif



		//Read two char - record type
		record_type = read_byte( input_file , record_number );
		if( record_type == -1)
			return -1;
		check_sum = check_sum + record_type;
		#ifdef DEBUG
		printf( "Record Type = %d\n", record_type );
		#endif


		//Set the old_address variable to address of first data record
		//Next time it will be overwritten by end address of block
		if( ( record_type == 0 ) && ( record_number == 1 ) ){
			old_address = address;
		}


		//Read data bytes
		switch(record_type){
			case 0:	//Read data
				#ifdef DEBUG
				printf("Data Bytes are: ");
				#endif
				for( i = 0 ; i < byte_count ; ++i ){
					int data = read_byte( input_file , record_number );
					if( data == -1)
						return -1;
					check_sum = check_sum + data;
					#ifdef DEBUG
					printf( "%02x ", data );
					#endif
				}
				#ifdef DEBUG
				printf("\n");
				#endif
				break;

			case 1: //End of file record
				record_end_flag = 1;
				break;

			case 2: //Extended Segment Address Record
				if( ( hex_type == 0 ) || ( hex_type == 2 ) ){
					hex_type = 2;
					temp1 = read_byte( input_file , record_number );
					if( temp1 == -1 )
						return -1;
					check_sum = check_sum + temp1;

					temp2 = read_byte( input_file , record_number );
					if( temp2 == -1 )
						return -1;
					check_sum = check_sum + temp2;
					es_address = (unsigned int)( (temp1 << 8 ) | temp2 );

					es_address = es_address << 4;
					extended_segment_addresses.push_back( es_address );
				}
				else{
					fprintf( stderr , "Invalid Record Type: Expected 2 but received %d\n" , record_type );
					return -1;
				}
				break;

			case 3: //Read CS|IP Register - used in Intel I16HEX
				if( ( hex_type == 0 ) || ( hex_type == 2 ) ){
					hex_type = 2;
					temp1 = read_byte( input_file , record_number );
					if( temp1 == -1 )
						return -1;
					check_sum = check_sum + temp1;

					temp2 = read_byte( input_file , record_number );
					if( temp2 == -1 )
						return -1;
					check_sum = check_sum + temp2;

					cs_register = (unsigned int)( (temp1 << 8 ) | temp2 );


					temp1 = read_byte( input_file , record_number );
					if( temp1 == -1 )
						return -1;
					check_sum = check_sum + temp1;

					temp2 = read_byte( input_file , record_number );
					if( temp2 == -1 )
						return -1;
					check_sum = check_sum + temp2;

					ip_register = (unsigned int)( (temp1 << 8 ) | temp2 );

					cs_register_values.push_back( cs_register );
					ip_register_values.push_back( ip_register );

				}
				else{
					fprintf( stderr , "Invalid Record Type: Expected 2 but received %d\n" , record_type );
					return -1;
				}
				break;

			case 4: //Extended Linear Address Record
				if( ( hex_type == 0 ) || ( hex_type == 4 ) ){
					hex_type = 4;
					temp1 = read_byte( input_file , record_number );
					if( temp1 == -1 )
						return -1;
					check_sum = check_sum + temp1;

					temp2 = read_byte( input_file , record_number );
					if( temp2 == -1 )
						return -1;
					check_sum = check_sum + temp2;

					lb_address = (unsigned int)( (temp1 << 8 ) | temp2 );

					lb_address = lb_address << 16;
					linear_base_addresses.push_back( lb_address );
				}
				else{
					fprintf( stderr , "Invalid Record Type: Expected 4 but received %d\n" , record_type );
					return -1;
				}
				break;

			case 5: //Read EIP register - used in Intel I32HEX
				if( ( hex_type == 0 ) || ( hex_type == 4 ) ){
					hex_type = 4;
					temp1 = read_byte( input_file , record_number );
					if( temp1 == -1 )
						return -1;
					check_sum = check_sum + temp1;

					temp2 = read_byte( input_file , record_number );
					if( temp2 == -1 )
						return -1;
					check_sum = check_sum + temp2;

					eip = (unsigned int)( (temp1 << 8 ) | temp2 );
					eip_values.push_back( eip );
				}
				else{
					fprintf( stderr , "Invalid Record Type: Expected 4 but received %d\n" , record_type );
					return -1;
				}
				break;

			defalut:
				fprintf( stderr , "Invalid record type in record number %d" , record_number );
				return -1;
		}




		//Check jump in address from one record to next
		if( record_type == 0 ) {
			//Store last address block and check if new address
			//has already beed repeated in previous blocks
			//Write code here
			if( hex_type == 0 ){
				//old_address = old_address;
				//new_address = new_address;
				absolute_address = address;
			}
			if( hex_type == 2 ) {
				//old_address = es_address + old_address;
				//new_address = es_address + new_address;
				absolute_address = es_address + address;
			}
			else if( hex_type == 4 ) {
				//old_address = lb_address + old_address;
				//new_address = lb_address + new_address;
				absolute_address = lb_address + address;
			}
		}
		if( ( record_type== 0) && ( ( absolute_address != new_address) && ( record_number > 1) ) ){
			if ( addrange( old_address , new_address ) == -1 )
				return -1;
			old_address = absolute_address;
		}
		if( record_type == 1 )
			addrange( old_address , new_address );



		//Read two char - check sum
		temp = read_byte( input_file , record_number );
		if( temp == -1 )
			return -1;

		check_sum = ( (check_sum & 0xff)  + temp ) & 0xff;
		if( check_sum == 0 ){
			if( record_type == 0 )
				new_address = absolute_address + byte_count;
		}
		else{
			fprintf( stderr , "Checksum of %d record number does not matches\n" , record_number );
			return -1;
		}

		#ifdef DEBUG
		printf( "Check sum = %x\n" , temp );
		#endif



		byteread =  parse_newline_character( input_file , record_number );
		if( byteread != 1 )
			return byteread;


		if(record_type == 0 )// || ( (record_type >= 2 ) && (record_type <= 5 ) )
			record_number += 1;


		#ifdef DEBUG
		printf("Record Number = %d\n" , record_number );
		#endif

		//Break while loop at the end of record
		//To cater the condition where file contains
		//either new line characters or anything else
		//after hitting last record (record_type = 2)
		if( record_end_flag == 1 )
			break;

	 }


	//Check only new line character should present
	//otherwise file is invalid.
	if( record_end_flag == 1 ) {
		while( 1 ) {
			byteread = read( input_file , &ch , 1 );
			if(byteread == 0) {
				//printf("End of file (After End of File)\n");
	 			return 1;
			}

			byteread =  parse_newline_character( input_file , record_number );
			if( byteread != 1 )
				return byteread;

			record_number += 1;
		}
	}

	return 0;
}


/*******************************************************************************
Function Name:main()
Description: parses the intel hex file and print information
Input: input file name
output: information about file
********************************************************************************/
int main(int argc, char **argv){

	int retVal;
	//unsigned int i;

	if(argc != 2){
		printf("Correct usage %s <input_file>\n",argv[0]);
		return 0;
	}

	long long file_len=0;
	struct stat sb;  	// get file length

	int input_file=0;
	
  	input_file = open(argv[1], O_RDONLY, "rb");
	if( input_file == -1){
		printf("ERR: Input file %s can not be open\n",argv[1]);
		return 0;
	}
	
	if (stat(argv[1], &sb) == -1) { 
		 	printf("\nERROR: Sysetm unable to read file length of %s \n",argv[1]); 
		 	return(-1);
	}
	file_len=sb.st_size;
	if(file_len == 0){
		printf("\nERROR: Input file is empty\n");
		return -1;
	}

	printf(" Parsing %s file\n",argv[1]);
	retVal = parse_file(input_file);

	if( ( retVal == 1 ) || ( ( retVal == -2) && (record_number > 1) ) ) {
		if( hex_type == 4 ){
			printf("Intel Hex-32 Format\n");
			printf("Linear Base Addresses are:\n" );
			for( unsigned int i=0 ; i < linear_base_addresses.size() ; ++i )
				printf("\t%x\n" , linear_base_addresses[i] );

			if( eip_values.size() > 0 ){
				printf("EIP register values are: \n" );
				for( unsigned int i=0 ; i < eip_values.size() ; ++i )
					printf("\t%x\n" , eip_values[i] );
			}
			else
				printf("EIP register not set\n");
		}
		else if( hex_type == 2 ){
			printf("Intel Hex-16 Format\n");
			printf("Extended Segment Addresses are: \n" );
			for( unsigned int i=0 ; i < extended_segment_addresses.size() ; ++i )
				printf("\t%x\n" , extended_segment_addresses[i] );

			if( cs_register_values.size() > 0 ) {
				printf("CS Register values are: \n" );
				for( unsigned int i=0 ; i < cs_register_values.size() ; ++i )
					printf("\t%x\n" , cs_register_values[i] );
			}
			else
				printf("CS Register is not set\n");

			if( ip_register_values.size() > 0 ) {
				printf("IP Register values are: \n" );
				for( unsigned int i=0 ; i < ip_register_values.size() ; ++i )
					printf("\t%x\n" , ip_register_values[i] );
			}
			else
				printf("IP Register is not set\n");
		}
		else{
			printf("Intel Hex-8 Format\n");
		}
		printf("Address Blocks:\n");
		for(unsigned int i=0 ; i<a.size() ; ++i )
			printf( "\tAddress block-%u: from %d  to %d\n" , i , a[i] , b[i] );
			
	}
	close(input_file);
	return 0;
}


